/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.dita;

import java.util.List;
import java.util.Set;

import javax.swing.text.BadLocationException;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.api.AuthorAccess;
import ro.sync.ecss.extensions.api.AuthorOperationException;
import ro.sync.ecss.extensions.api.ContentInterval;
import ro.sync.ecss.extensions.api.node.AttrValue;
import ro.sync.ecss.extensions.api.node.AuthorElement;
import ro.sync.ecss.extensions.api.node.AuthorNode;
import ro.sync.ecss.extensions.api.table.operations.AuthorTableDeleteColumnArguments;
import ro.sync.ecss.extensions.api.table.operations.AuthorTableDeleteRowArguments;
import ro.sync.ecss.extensions.api.table.operations.AuthorTableInsertColumnArguments;
import ro.sync.ecss.extensions.api.table.operations.AuthorTableOperationsHandler;
import ro.sync.ecss.extensions.api.table.operations.TableColumnSpecificationInformation;
import ro.sync.ecss.extensions.commons.table.operations.TableOperationsUtil;
import ro.sync.ecss.extensions.commons.table.operations.cals.CALSTableColumnSpecificationInformation;
import ro.sync.ecss.extensions.commons.table.support.CALSColSpec;
import ro.sync.ecss.extensions.commons.table.support.CALSTableCellInfoProvider;
import ro.sync.ecss.extensions.dita.topic.table.InsertTableOperation;

/**
 * Author table operations handler for DITA framework.
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class DITAAuthorTableOperationsHandler extends AuthorTableOperationsHandler {

  /**
   * Insert row operation for DITA simpletable
   */
  private ro.sync.ecss.extensions.dita.topic.table.simpletable.InsertRowOperation insertRowSimpleTable = null;
  /**
   * Insert column operation for DITA simpletable
   */
  private ro.sync.ecss.extensions.dita.topic.table.simpletable.InsertColumnOperation insertColumnSimpleTable = null;
  /**
   * Insert row operation for DITA CALS table
   */
  private ro.sync.ecss.extensions.commons.table.operations.cals.InsertRowOperation insertRowCALSTable = null;
  /**
   * Insert column operation for DITA CALS table
   */
  private ro.sync.ecss.extensions.commons.table.operations.cals.InsertColumnOperation insertColumnCALSTable = null;
  
  /**
   * Delete column operation for DITA simpletable
   */
  private ro.sync.ecss.extensions.dita.topic.table.simpletable.DeleteColumnOperation deleteColumnSimpleTable = null;
  
  /**
   * Delete column operation for DITA CALS table
   */
  private ro.sync.ecss.extensions.commons.table.operations.cals.DeleteColumnOperation deleteColumnCALSTable = null;
  
  /**
   * Delete row operation for DITA simpletable
   */
  private ro.sync.ecss.extensions.dita.topic.table.simpletable.DeleteRowOperation deleteRowSimpleTable = null;
  
  /**
   * Delete row operation for DITA CALS table
   */
  private ro.sync.ecss.extensions.commons.table.operations.cals.DeleteRowOperation deleteRowCALSTable = null;
  
  /**
   * Insert table operation.
   */
  private InsertTableOperation insertTableOperation = null;
  
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(DITAAuthorTableOperationsHandler.class
      .getName());
  
  /**
   * @see ro.sync.ecss.extensions.api.table.operations.AuthorTableOperationsHandler#handleInsertColumn(ro.sync.ecss.extensions.api.table.operations.AuthorTableInsertColumnArguments)
   */
  @Override
  public boolean handleInsertColumn(AuthorTableInsertColumnArguments tablePasteColumnArgs) throws AuthorOperationException {
    boolean handled = false;
    try {
      AuthorAccess authorAccess = tablePasteColumnArgs.getAuthorAccess();
      boolean simpleTable = isSimpleTable(authorAccess, tablePasteColumnArgs.getInsertOffset());
      
      if (insertTableOperation == null) {
        insertTableOperation = new InsertTableOperation();
      }

      // Perform operation
      if (simpleTable) {
        if (insertRowSimpleTable == null) {
          // Initialize simpletable insert row operation
          insertRowSimpleTable = new ro.sync.ecss.extensions.dita.topic.table.simpletable.InsertRowOperation();
        }
        if (insertColumnSimpleTable == null) {
          // Initialize simpletable insert column operation
          insertColumnSimpleTable = new ro.sync.ecss.extensions.dita.topic.table.simpletable.InsertColumnOperation();
        }
        
        // Insert column
        insertColumnSimpleTable.performInsertColumn(
            authorAccess, null, tablePasteColumnArgs.getColumnFragments(), 
            tablePasteColumnArgs.getColumnSpecificationInformation(),
            tablePasteColumnArgs.areFragmentsWrappedInCells(), 
            insertRowSimpleTable, insertTableOperation);
      } else {
        if (insertRowCALSTable == null) {
          // Initialize CALS insert row operation
          insertRowCALSTable = new ro.sync.ecss.extensions.commons.table.operations.cals.InsertRowOperation();
        }
        if (insertColumnCALSTable == null) {
          // Initialize CALS insert column operation
          insertColumnCALSTable = new ro.sync.ecss.extensions.commons.table.operations.cals.InsertColumnOperation();
        }
        
        // Insert column
        insertColumnCALSTable.performInsertColumn(
            authorAccess, null, tablePasteColumnArgs.getColumnFragments(), 
            tablePasteColumnArgs.getColumnSpecificationInformation(),
            tablePasteColumnArgs.areFragmentsWrappedInCells(), 
            insertRowCALSTable, insertTableOperation);
      }
      // The paste operation was handled
      handled = true;
    } catch (IllegalArgumentException e) {
      // Nothing to do
      logger.warn("Paste column failed.", e);
    } catch (BadLocationException e) {
      // Nothing to do
      logger.warn("Paste column failed.", e);
    }
    return handled;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.table.operations.AuthorTableOperationsHandler#handleDeleteColumn(ro.sync.ecss.extensions.api.table.operations.AuthorTableDeleteColumnArguments)
   */
  @Override
  public boolean handleDeleteColumn(AuthorTableDeleteColumnArguments arguments)
  throws AuthorOperationException {
    boolean handled = false;
    try {
      AuthorAccess authorAccess = arguments.getAuthorAccess();
      
      int tableOffset = -1;
      List<ContentInterval> intervals = arguments.getColumnCellsIntervals();
      if (intervals != null && intervals.size() > 0) {
        tableOffset = intervals.get(0).getEndOffset();
      }
      
      if (tableOffset >= 0) {
        boolean simpleTable = isSimpleTable(authorAccess, tableOffset);

        // Perform operation
        if (simpleTable) {
          if (deleteColumnSimpleTable == null) {
            // simpletable delete column operation
            deleteColumnSimpleTable = new ro.sync.ecss.extensions.dita.topic.table.simpletable.DeleteColumnOperation();
          }
          deleteColumnSimpleTable.performDeleteColumn(authorAccess, arguments.getColumnCellsIntervals(), false);
        } else {
          if (deleteColumnCALSTable == null) {
            // Initialize CALS delete column operation
            deleteColumnCALSTable = new ro.sync.ecss.extensions.commons.table.operations.cals.DeleteColumnOperation();
          }
          deleteColumnCALSTable.performDeleteColumn(authorAccess, arguments.getColumnCellsIntervals(), false);
        }
        handled = true;
      }
    } catch (IllegalArgumentException e) {
      // Nothing to do
      logger.warn("Delete column failed.", e);
    } catch (BadLocationException e) {
      // Nothing to do
      logger.warn("Delete column failed.", e);
    }
    return handled;
  }
  
  /**
   * Check if the current table is a simple table.
   * 
   * @param authorAccess The author access.
   * @param tableOffset The table offset.
   * 
   * @return <code>true</code> if the current table is a simple table.
   * 
   * @throws BadLocationException
   */
  private boolean isSimpleTable(AuthorAccess authorAccess, int tableOffset) throws BadLocationException {
    AuthorNode currentNode = authorAccess.getDocumentController().getNodeAtOffset(tableOffset);
    boolean simpleTable = false;
    while (currentNode != authorAccess.getDocumentController().getAuthorDocumentNode()) {
      if (currentNode instanceof AuthorElement) {
        AuthorElement pasteElement = (AuthorElement) currentNode;
        AttrValue classAttribute = pasteElement.getAttribute("class");
        if (classAttribute != null) {
          String classValue = classAttribute.getRawValue();
          if (classValue != null) {
            if (classValue.contains(" topic/strow ") || 
                classValue.contains(" topic/sthead ") ||
                classValue.contains(" topic/stentry ") ||
                classValue.contains(" topic/simpletable ")) {
              // The paste offset is in simpletable context
              simpleTable = true;
              break;
            } else  if (classValue.contains(" topic/row ") || 
                classValue.contains(" topic/head ") ||
                classValue.contains(" topic/entry ") ||
                classValue.contains(" topic/table ")) {
              // The paste offset is in CALS table context
              break;
            }
          }
        }
      }

      currentNode = currentNode.getParent();
    }
    
    return simpleTable;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.table.operations.AuthorTableOperationsHandler#handleDeleteRow(ro.sync.ecss.extensions.api.table.operations.AuthorTableDeleteRowArguments)
   */
  @Override
  public boolean handleDeleteRow(AuthorTableDeleteRowArguments arguments)
  throws AuthorOperationException {
    boolean handled = false;
    try {
      AuthorAccess authorAccess = arguments.getAuthorAccess();
      ContentInterval rowInterval = arguments.getRowInterval();
      AuthorNode rowNode = authorAccess.getDocumentController().getNodeAtOffset(rowInterval.getStartOffset() + 1);
      if (rowNode.getStartOffset() == rowInterval.getStartOffset() && rowNode.getEndOffset() == rowInterval.getEndOffset() - 1) {
        boolean simpleTable = isSimpleTable(authorAccess, rowInterval.getStartOffset());

        // Perform operation
        if (simpleTable) {
          if (deleteRowSimpleTable == null) {
            // simpletable delete column operation
            deleteRowSimpleTable = new ro.sync.ecss.extensions.dita.topic.table.simpletable.DeleteRowOperation();
          }
          deleteRowSimpleTable.performDeleteRows(authorAccess, rowInterval.getStartOffset(), rowInterval.getEndOffset());
        } else {
          if (deleteRowCALSTable == null) {
            // Initialize CALS delete column operation
            deleteRowCALSTable = new ro.sync.ecss.extensions.commons.table.operations.cals.DeleteRowOperation();
          }
          deleteRowCALSTable.performDeleteRows(authorAccess, rowInterval.getStartOffset(), rowInterval.getEndOffset());
        }
        handled = true;
      } else {
        throw new AuthorOperationException(
            "The operation cannot be performed due to: wrong row content interval argument");
      }
    } catch (IllegalArgumentException e) {
      // Nothing to do
      logger.warn("Delete row failed.", e);
    } catch (BadLocationException e) {
      // Nothing to do
      logger.warn("Delete row failed.", e);
      throw new AuthorOperationException(
          "The operation cannot be performed due to: " + e.getMessage(), e);
    }
    return handled;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.table.operations.AuthorTableOperationsHandler#getTableElementContainingOffset(ro.sync.ecss.extensions.api.AuthorAccess, int)
   */
  @Override
  public AuthorElement getTableElementContainingOffset(AuthorAccess access, int offset) {
    String[] tableClassValues = new String[] {
        " topic/simpletable ", 
        " topic/table "
    };
    // Find the table element
    return TableOperationsUtil.getTableElementContainingOffset(offset,  
        access, tableClassValues);
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.table.operations.AuthorTableOperationsHandler#getColumnSpecification(ro.sync.ecss.extensions.api.AuthorAccess, ro.sync.ecss.extensions.api.node.AuthorElement, int)
   */
  @Override
  public TableColumnSpecificationInformation getColumnSpecification(AuthorAccess authorAccess, 
      AuthorElement tableElement, int columnIndex) {
    TableColumnSpecificationInformation colspecInfo = null;
    Boolean isSimpleTableElem = null;
    try {
      // Determin if the given table is a simpletable or a CALS table
      isSimpleTableElem = isSimpleTable(authorAccess, tableElement.getStartOffset() + 1);
    } catch (BadLocationException e) {
      logger.error(e);
    }
    if (isSimpleTableElem != null && !isSimpleTableElem) {
      // CALS table
      CALSTableCellInfoProvider cellInfoProvider = new CALSTableCellInfoProvider();
      cellInfoProvider.init(tableElement);
      // Determine the column specification 
      Set<CALSColSpec> colSpecs = cellInfoProvider.getColSpecs();
      if (colSpecs.size() > columnIndex) {
        // Found the column specification
        CALSColSpec colSpec = colSpecs.toArray(new CALSColSpec[0])[columnIndex];
        if (colSpec != null) {
          colspecInfo = new CALSTableColumnSpecificationInformation(
              colSpec.getColWidth(), colSpec.getColumnName());
        }
      }
    }
    return colspecInfo;
  }
}
