/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2009 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.dita;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.dita.DITAAccess;
import ro.sync.ecss.extensions.api.link.ElementLocator;
import ro.sync.ecss.extensions.api.link.IDTypeVerifier;
import ro.sync.ecss.extensions.commons.DefaultElementLocatorProvider;

/**
 * Implementation for locating elements based on a link from a DITA document.
 * 
 *  See:
 *  <br>
 *  <a href="http://docs.oasis-open.org/dita/v1.0/langspec/relatedl.html">
 *  http://docs.oasis-open.org/dita/v1.0/langspec/relatedl.html
 *  </a>
 *  <br>
 *  <a href="http://docs.oasis-open.org/dita/v1.1/OS/langspec/common/theconrefattribute.html">
 *  http://docs.oasis-open.org/dita/v1.1/OS/langspec/common/theconrefattribute.html
 *  </a>
 *  <br>
 *  <a href="http://docs.oasis-open.org/dita/v1.0/langspec/xref.html">
 *  http://docs.oasis-open.org/dita/v1.0/langspec/xref.html
 *  </a>
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class DITAElementLocatorProvider extends DefaultElementLocatorProvider {
  /**
   * @see ro.sync.ecss.extensions.commons.DefaultElementLocatorProvider#getElementLocator(ro.sync.ecss.extensions.api.link.IDTypeVerifier, java.lang.String)
   */
  @Override
  public ElementLocator getElementLocator(IDTypeVerifier idVerifier, String link) {
    ElementLocator locator = null;
    if (link.startsWith("element(")) {
      // Link type is XInclude element scheme.
      locator = super.getElementLocator(idVerifier, link);
    } else if (link.startsWith("keyref(") && link.endsWith(")")) {
      // EXM-20978 Link to a key definition in a DITA Map.
     locator = new DITAMapKeyDefElementLocator(link);
    } else {
      boolean keyRef = link.startsWith(DITAAccess.ID_ANY + "/");
      if (keyRef) {
        link = link.substring((DITAAccess.ID_ANY + "/").length(), link.length());
      }
      if (link.indexOf('/') != -1) {
        if (link.indexOf('/') != link.length() - 1) {
          // This is a link to a non topic element. 
          // The parent topic ID and the id of the referred element  are specified.
          locator = new DITAElementLocator(link);
        } 
      } else {
        // This a link to an ID type attribute.
        locator = new DITAIDElementLocator(
            //EXM-15220 For DITA the "id" attribute is not really defined as an ID, be more flexible.
            new IDTypeVerifier() {
              public boolean hasIDType(String elementName, String elementNs, String attrName, String attrNs) {
                return "id".equals(attrName);
              }
            }, link, keyRef);
      }
    }
    
    return locator;
  }
}