/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.dita.map.topicref;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.LinkedHashMap;

import javax.xml.transform.sax.SAXSource;

import org.apache.log4j.Logger;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.XMLReader;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.dita.DITAAccess;
import ro.sync.ecss.dita.Reference;
import ro.sync.ecss.dita.reference.keyref.KeyInfo;
import ro.sync.ecss.extensions.api.AuthorAccess;
import ro.sync.ecss.extensions.api.DITAMapReferencesResolver;
import ro.sync.ecss.extensions.api.ValidatingReferenceResolverException;
import ro.sync.ecss.extensions.api.node.AttrValue;
import ro.sync.ecss.extensions.api.node.AuthorDocument;
import ro.sync.ecss.extensions.api.node.AuthorElement;
import ro.sync.ecss.extensions.api.node.AuthorNode;
import ro.sync.ecss.extensions.dita.conref.DITAConRefResolver;
import ro.sync.ecss.extensions.dita.conref.DITAXMLReaderWrapper;
import ro.sync.exml.workspace.api.util.XMLReaderWithGrammar;

/**
 * Resolves the hrefs to other maps.
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class DITAMapRefResolver extends DITAConRefResolver implements DITAMapReferencesResolver{
  
  /**
   * Logger for logging. 
   */
  private static Logger logger = Logger.getLogger(DITAMapRefResolver.class.getName());
  /**
   * The <code>keyref</code> attribute.
   */
  private static final String ATTRIBUTE_KEYREF = "keyref";
  /**
   * The <code>href</code> attribute.
   */
  private static final String ATTRIBUTE_HREF = "href";
  
  /**
   * The <code>scope</code> attribute.
   */
  private static final String ATTRIBUTE_SCOPE = "scope";
  
  /**
   * The <code>format</code> attribute.
   */
  private static final String ATTRIBUTE_FORMAT = "format";
  
  /**
   * The <code>format</code> attribute.
   */
  private static final String ATTRIBUTE_CLASS = "class";
  
  /**
   * The <code>processing-role</code> attribute.
   */
  private static final String ATTRIBUTE_PROCESSING_ROLE = "processing-role";

  /**
   *  If true, will resolve both map references and topic references.
   * If false, will resolve only map references, defaults to false
   */
  private boolean resolveAllTopicRefs = false;
  
  /**
   * An element that has <code>href</code> attribute 
   * is an element with references.
   * 
   * @see ro.sync.ecss.extensions.api.AuthorReferenceResolver#hasReferences(ro.sync.ecss.extensions.api.node.AuthorNode)
   */
  @Override
  public boolean hasReferences(AuthorNode node) {
    boolean hasReferences = getHref(node) != null;
    if(! hasReferences) {
      hasReferences = super.hasReferences(node);
    }
    return hasReferences;
  }
  
  /**
   * Check if a node has the keyref attribute.
   * 
   * @param node The Author node.
   * @return <code>true</code> if the given node has the keyref attribute
   */
  private boolean isKeyReference(AuthorNode node) {
    boolean keyRef = false;
    if (node.getType() == AuthorNode.NODE_TYPE_ELEMENT) {
      AuthorElement element = (AuthorElement) node;
      AttrValue classValue = element.getAttribute(ATTRIBUTE_CLASS);
      // Check the class value
      if (classValue != null && classValue.getValue().contains(" map/topicref ")) {
        // Check if the keyref attribute is declared
        AttrValue keyrefValue = element.getAttribute(ATTRIBUTE_KEYREF);
        if (keyrefValue != null) {
          keyRef = true;
        }
      }
    }
    return keyRef;
  }

  /**
   * Get the href value for a node.
   * @param node The node
   * @return the href value or null.
   */
  private String getHref(AuthorNode node) {
    String href = null;
    if (node.getType() == AuthorNode.NODE_TYPE_ELEMENT) {
      AuthorElement element = (AuthorElement) node;
      AttrValue classValue = element.getAttribute(ATTRIBUTE_CLASS);
      if (classValue != null && classValue.getValue().contains(" map/topicref ")) {
        String hrefValueStr = null;
        AttrValue hrefValue = element.getAttribute(ATTRIBUTE_HREF);
        if(hrefValue != null) {
          //We have a href
          hrefValueStr = hrefValue.getValue();
        } else {
          //EXM-17624 Expand topicref keyrefs.
          //Maybe we have a keyref, get the href from it instead
          AttrValue keyrefValue = element.getAttribute(ATTRIBUTE_KEYREF);
          if (keyrefValue != null) {
            String keyrefStrValue = keyrefValue.getValue();
            LinkedHashMap<String, KeyInfo> keys = DITAAccess.getKeys(node.getXMLBaseURL());
            if (keys != null) {
              //Resolve the key reference.
              KeyInfo keyInfo = keys.get(keyrefStrValue);
              if (keyInfo != null && keyInfo.getHrefLocation() != null) {
                hrefValueStr = keyInfo.getHrefLocation().toString();
              }
            }
          }
        }
        if (hrefValueStr != null) {
          AttrValue scopeValue = element.getAttribute(ATTRIBUTE_SCOPE);
          if (scopeValue == null || !"external".equals(scopeValue.getValue())) {
            boolean resolve = false;
            boolean isTopicRef = false;
            boolean isMapRef = false;
            AttrValue formatValue = element.getAttribute(ATTRIBUTE_FORMAT);
            if (formatValue == null) {
              isTopicRef = true;
            } else {
              String formatStr = formatValue.getValue();
              if ("dita".equals(formatStr)) {
                isTopicRef = true;  
              } else if("ditamap".equals(formatStr)) {
                isMapRef = true;  
              }
            }
            
            if (isMapRef || isTopicRef) {
              if (resolveAllTopicRefs) {
                //Resolve all references which are not in rell tables
                resolve = !isInRelCell(node);
              } else {
                resolve = isMapRef;
              }
            }
            if (resolve) {
              href = hrefValueStr;
            }
          }
        }
      }
    }
    return href;
  }
  
  /**
   * Check if a node is in a rell cell
   * @param node The node
   * @return true if the node is in a rel cell
   */
  private boolean isInRelCell(AuthorNode node) {
    String relcell = " map/relcell ";
    while(node != null) {
      if(node.getType() == AuthorNode.NODE_TYPE_ELEMENT) {
        AuthorElement elem = (AuthorElement) node;
        AttrValue classAttr = elem.getAttribute(ATTRIBUTE_CLASS);
        if(classAttr != null && classAttr.getValue().contains(relcell)) {
          return true;
        }
      }
      node = node.getParent();
    }
    return false;
  }
  
  /**
   * Returns the value of the <code>href</code> attribute.
   * 
   * @see ro.sync.ecss.extensions.api.AuthorReferenceResolver#getDisplayName(ro.sync.ecss.extensions.api.node.AuthorNode)
   */
  @Override
  public String getDisplayName(AuthorNode node) {
    String displayName = getHref(node);
    if(displayName == null) {
      return super.getDisplayName(node);
    }
    return displayName;
  }
  
  /**
   * Caches access to the grammar pool
   */
  private Object grammarCache = null;
  
  /**
   * Resolve the content referred by <code>conref</code> attribute.
   * 
   * @see ro.sync.ecss.extensions.api.AuthorReferenceResolver#resolveReference(ro.sync.ecss.extensions.api.node.AuthorNode, java.lang.String, AuthorAccess, EntityResolver)
   */
  @Override
  public SAXSource resolveReference(
      AuthorNode node, 
      String systemID, 
      AuthorAccess authorAccess,
      EntityResolver  entityResolver) {
    String href = getHref(node);
    if(href != null) {
      SAXSource saxSource = null;

      try {
        //Parse the reference
        Reference conRef = DITAAccess.parseDITAHref(systemID, href);
        if (conRef != null) {
          //Resolve the entity through catalog
          InputSource inputSource = entityResolver.resolveEntity(null, conRef.getUri());
          if(inputSource == null) {
            inputSource = new InputSource(conRef.getUri());
          }

          //EXM-16083 XML Reader with grammar cache
          XMLReaderWithGrammar readerWithCache = authorAccess.getXMLUtilAccess().newNonValidatingXMLReader(grammarCache);
          XMLReader xmlReader = readerWithCache.getXmlReader();
          grammarCache = readerWithCache.getGrammarCache();
          xmlReader.setEntityResolver(entityResolver);

          String[] topicPath = getTopicPath(conRef.getTopicID());            
          //Create the XML reader wrapper if this is the case.
          xmlReader = 
            new DITAXMLReaderWrapper(
                xmlReader, topicPath, null, isKeyReference(node));
          saxSource = new SAXSource(xmlReader, inputSource);
        }
      } catch (MalformedURLException e) {
        logger.error(e, e);
      } catch (SAXNotRecognizedException e) {
        logger.error(e, e);
      } catch (SAXNotSupportedException e) {
        logger.error(e, e);
      } catch (SAXException e) {
        logger.error(e, e);
      } catch (IOException e) {
        logger.error(e, e);
      } 
      
      return saxSource;
    } else {
      //Delegate to super for conref.
      return super.resolveReference(node, systemID, authorAccess, entityResolver);
    }
  }
  
  /**
   * Get the reference System ID
   * 
   * @see ro.sync.ecss.extensions.api.AuthorReferenceResolver#getReferenceSystemID(ro.sync.ecss.extensions.api.node.AuthorNode, AuthorAccess)
   */
  @Override
  public String getReferenceSystemID(AuthorNode node, AuthorAccess authorAccess) {
    String href = getHref(node);
    if (href != null) {
      String systemID = null;
      try {
        //Create the absolute system ID
        systemID = new URL(node.getXMLBaseURL(), href).toString();

        //Try to pass through catalog.
        Reference parseDITAHref = DITAAccess.parseDITAHref(
            node.getXMLBaseURL()  == null ? null : node.getXMLBaseURL().toString(), 
                href);
        if(parseDITAHref != null) {
          systemID = parseDITAHref.getUri();
          if(parseDITAHref.getTopicID() != null && parseDITAHref.getTopicID().length() > 0) {
            systemID +="#" + parseDITAHref.getTopicID();
          }
        }
      } catch (MalformedURLException e) {
        if (logger.isDebugEnabled()) {
          logger.debug(e, e);
        }
      }
      return systemID;
    } else {
      return super.getReferenceSystemID(node, authorAccess);
    }
  }
  
  /**
   * 
   * @see ro.sync.ecss.extensions.api.ValidatingAuthorReferenceResolver#checkTarget(ro.sync.ecss.extensions.api.node.AuthorNode, ro.sync.ecss.extensions.api.node.AuthorDocument)
   */
  @Override
  public void checkTarget(AuthorNode node, AuthorDocument targetDocument) throws ValidatingReferenceResolverException{
    if(getHref(node) != null) {
      //Do not validate source and target, topic reference.
    } else {
      //Maybe conref
      super.checkTarget(node, targetDocument);
    }
  }
  
  /**
   * The value of <code>conref</code> attribute is used as the unique identifier.
   * 
   * @see ro.sync.ecss.extensions.api.AuthorReferenceResolver#getReferenceUniqueID(ro.sync.ecss.extensions.api.node.AuthorNode)
   */
  @Override
  public String getReferenceUniqueID(AuthorNode node) {
    String id = getHref(node);
    if(id == null) {
      //Maybe conref
      return super.getReferenceUniqueID(node);
    }
    return id;
  }

  /**
   * Returns <code>true</code> when the attribute name is equal to <code>'conref'</code>.
   * 
   * @see ro.sync.ecss.extensions.api.AuthorReferenceResolver#isReferenceChanged(ro.sync.ecss.extensions.api.node.AuthorNode, java.lang.String)
   */
  @Override
  public boolean isReferenceChanged(AuthorNode node, String attributeName) {
    //Check for href or delegate for conref.
    return ATTRIBUTE_HREF.equals(attributeName) 
        || ATTRIBUTE_SCOPE.equals(attributeName)
        || ATTRIBUTE_FORMAT.equals(attributeName)
        || ATTRIBUTE_PROCESSING_ROLE.equals(attributeName)
        || super.isReferenceChanged(node, attributeName);
  }

  /**
   * @see ro.sync.ecss.extensions.api.DITAMapReferencesResolver#setResolveAllTopicReferences(boolean)
   */
  public void setResolveAllTopicReferences(boolean resolveAllTopicRefs) {
    this.resolveAllTopicRefs = resolveAllTopicRefs;
  }
}
