/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.exml.workspace.api.editor;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.exml.editor.EditorPageConstants;
import ro.sync.exml.workspace.api.editor.page.WSEditorPage;
import ro.sync.exml.workspace.api.editor.page.author.WSAuthorEditorPage;
import ro.sync.exml.workspace.api.editor.page.text.WSTextEditorPage;
import ro.sync.exml.workspace.api.editor.page.text.xml.WSXMLTextEditorPage;
import ro.sync.exml.workspace.api.editor.validation.ValidationProblemsFilter;
import ro.sync.exml.workspace.api.listeners.WSEditorListener;
import ro.sync.exml.workspace.api.listeners.WSEditorPageChangedListener;


/**
 * Provides access to methods related to the editor actions and information.
 * 
 * @since 11.2 
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PUBLIC)
public interface WSEditor extends WSEditorBase, EditorPageConstants{
  /**
   * Get access to the current page.
   * @return the current page access. Can be null for pages which do not have special access methods
   * (for example Grid or Design).
   * <br>
   * For the Text page this return an implementation of {@link WSTextEditorPage}. 
   * If the Text page is for an XML Editor, this return an implementation of {@link WSXMLTextEditorPage}. 
   * <br>
   * For the Author page this return an implementation of {@link WSAuthorEditorPage}.
   */
  WSEditorPage getCurrentPage();
  
  /**
   * Get the ID of the current page.
   * 
   * @return The ID of the page, one of the constant fields: 
   * {@link EditorPageConstants#PAGE_TEXT}, {@link EditorPageConstants#PAGE_AUTHOR}, {@link EditorPageConstants#PAGE_GRID}, 
   * {@link EditorPageConstants#PAGE_DESIGN}, {@link EditorPageConstants#PAGE_DITA_MAP}
   */
  public String getCurrentPageID();
  
  /**
   * Add a listener for page changed events.
   * @since 12
   * @param pageChangedListener The page changed listener.
   */
  public void addPageChangedListener(WSEditorPageChangedListener pageChangedListener);
  
  /**
   * Remove the listener for page changed events.
   * @since 12
   * @param pageChangedListener The page changed listener.
   */
  public void removePageChangedListener(WSEditorPageChangedListener pageChangedListener);
  
  /**
   * Add a listener for editor related events.
   * 
   * @param editorListener The editor listener.
   * @since 13
   */
  public void addEditorListener(WSEditorListener editorListener);
  
  /**
   * Remove the listener for editor events.
   * 
   * @param editorListener The editor listener.
   * @since 13
   */
  public void removeEditorListener(WSEditorListener editorListener);
  
  /**
   * Change the current selected page in the editor. This does not affect editors opened in the DITA Maps Manager.
   * If problems occur during the page switch or the page ID is not recognized the page will be switched to Text and the operation is aborted.
   * 
   * @param pageID The ID of the page, one of the constant fields: 
   * {@link EditorPageConstants#PAGE_TEXT}, {@link EditorPageConstants#PAGE_AUTHOR}, {@link EditorPageConstants#PAGE_GRID}, 
   * {@link EditorPageConstants#PAGE_DESIGN}
   * 
   * @since 13
   */
  public void changePage(String pageID);
  
  /**
   * Add a filter for problems encountered during validation of the current editor. Validation can be manual or automatic.
   * Automatic validation is done when modifications occur in the XML file.
   * 
   * @param validationProblemsFilter a filter for problems encountered during validation of the current editor.
   * 
   * @since 13
   */
  public void addValidationProblemsFilter(ValidationProblemsFilter validationProblemsFilter);
  
  /**
   * Remove a filter for problems encountered during validation of the current editor. Validation can be manual or automatic.
   * Automatic validation is done when modifications occur in the XML file.
   * 
   * @param validationProblemsFilter a filter for problems encountered during validation of the current editor.
   * 
   * @since 13
   */
  public void removeValidationProblemsFilter(ValidationProblemsFilter validationProblemsFilter);
}